use std::{fs::File, io::Write, os::unix::prelude::PermissionsExt, path::Path, process::Command};

use anyhow::Context;
use reqwest::{blocking::Client, StatusCode};
use serde::Deserialize;

use crate::config;

const PRYMN_PATH: &str = "/usr/local/bin/prymn_agent";

pub fn install(token: &str) -> anyhow::Result<()> {
    let this_exe = std::env::current_exe()?;

    copy_binary(&this_exe, Path::new(PRYMN_PATH)).with_context(|| {
        format!(
            "could not copy the file {} to the destination {PRYMN_PATH}",
            this_exe.to_str().unwrap(),
        )
    })?;

    install_service_file(Path::new("/etc/systemd/system/prymn.service"))
        .context("could not install the agent daemon service")?;

    register_to_backend(token).context("could not register the agent")?;

    Ok(())
}

fn copy_binary(src: &Path, dest: &Path) -> anyhow::Result<()> {
    if dest.exists() {
        // unlink the potentially running binary
        std::fs::remove_file(dest)?;
    }

    std::fs::copy(src, dest)?;

    let mut perms = dest.metadata()?.permissions();
    perms.set_mode(0o755);
    std::fs::set_permissions(dest, perms)?;

    Ok(())
}

fn install_service_file(dest: &Path) -> anyhow::Result<()> {
    let mut file = File::create(dest)?;
    write!(
        file,
        r#"
[Unit]
Description=Prymn Agent Service
After=network.target

[Service]
ExecStart={PRYMN_PATH}
Type=simple
Restart=always

[Install]
WantedBy=default.target
"#
    )?;

    if !Command::new("systemctl")
        .arg("daemon-reload")
        .status()?
        .success()
    {
        anyhow::bail!("command exit with non-zero exit code; could not reload systemd daemon");
    }

    if !Command::new("systemctl")
        .arg("enable")
        .arg("--now")
        .arg("prymn.service")
        .status()?
        .success()
    {
        anyhow::bail!("command exit with non-zero exit code; could not enable systemd service");
    }

    Ok(())
}

fn register_to_backend(token: &str) -> anyhow::Result<()> {
    let client = Client::new();
    let response = client
        .post(format!(
            "{}/api/v1/servers/register",
            config::CONFIG.backend_url
        ))
        .json(&serde_json::json!({ "token": token }))
        .send()?;

    // TODO: When the backend API is established more concretely, change this to something better.
    #[derive(Deserialize)]
    struct ApiError {
        errors: serde_json::Value,
    }

    match response.status() {
        StatusCode::UNPROCESSABLE_ENTITY => {
            let error = response.json::<ApiError>()?;
            anyhow::bail!(
                "request was unsuccessful: the backend received invalid data: {}",
                error.errors.to_string()
            )
        }
        status if !status.is_success() => {
            anyhow::bail!("request was unsuccessful: error {}", status)
        }
        _ => Ok(()),
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn copy_binary_works() {
        let temp_dir = std::env::temp_dir();
        // let temp_dir = tempdir().unwrap();
        let file1_path = temp_dir.join("file1");
        let mut file1 = File::create(&file1_path).unwrap();
        let file2_path = temp_dir.join("file2");
        let mut file2 = File::create(&file2_path).unwrap();

        writeln!(file1, "old data").unwrap();
        writeln!(file2, "new data").unwrap();

        copy_binary(&file2_path, &file1_path).expect("could not copy file");

        let perms = file1_path
            .metadata()
            .expect("could not retrieve metadata")
            .permissions();

        let new_data = std::fs::read_to_string(file1_path).unwrap();

        assert_eq!(new_data, "new data\n");
        assert!(perms.mode() & 0o755 == 0o755);
    }
}
